#include <stdio.h>
#include <math.h>
#include <limits.h>
#include <iostream>  
#include <math.h> 
#include <Eigen/Dense>
#include <Eigen/LU>
#include <Eigen/Cholesky>
#include <random>
#include <exception>
#include <vector>
#include <nlopt.hpp>

#include "kriging.h"
#include "QAP.h"
#include "InOut.h"
#include "ants.h"
#include "utilities.h"
#include "timer.h"


using namespace Eigen;
using namespace std;

struct my_data data;
int eval_count;
double eval_min, eval_sol;

double e_theta, e_ssq, e_mu;
Eigen::VectorXd e_yon;
Eigen::MatrixXd e_psi;
Eigen::MatrixXd e_psinv; 



void removeDistance(int removeIndex){
    int n_sol=data.distance_matrix.cols();
  
    //row removal
    data.distance_matrix.block(removeIndex, 0, n_sol-removeIndex-1, n_sol) = data.distance_matrix.block(removeIndex+1, 0, n_sol-removeIndex-1, n_sol).eval();
    data.distance_matrix.block(0, removeIndex, n_sol, n_sol-removeIndex-1) = data.distance_matrix.block(0, removeIndex+1, n_sol, n_sol-removeIndex-1).eval();

    data.distance_matrix.conservativeResize(n_sol-1, n_sol-1);

}

MatrixXd CorrGauss(double theta, MatrixXd& distance){
    MatrixXd result = distance * (-theta);
    result=result.array().exp();
    return(result);
}

double objective(const std::vector<double> &x, std::vector<double> &grad, void *my_func_data){
    my_data *data= (my_data*) my_func_data;
    long int n_sol = data->evaluations.size();

    eval_count++;
    
    e_theta = pow(10, x[0]);
    e_psi = CorrGauss(e_theta, data->distance_matrix);
    e_psinv.resize(n_sol,n_sol);

    double psi_det;
    
    Eigen::FullPivLU<MatrixXd> lu(e_psi);
    if(lu.isInvertible()){
     /*
       psi_det = lu.determinant();
       psi_det = abs(psi_det);
       e_psiinv = lu.inverse();*/
       LLT<MatrixXd> psi1 =e_psi.llt();
       e_psinv = MatrixXd::Identity(n_sol,n_sol);
       e_psinv = psi1.solve(e_psinv).eval();
       psi_det = lu.determinant();
       psi_det = abs(psi_det);
    }else{
       cout << "Not invertible matrix found.\n";
       return(std::numeric_limits<double>::max());
    }
     
    double psi_sum = e_psinv.sum();
    if(psi_sum==0){
        return(std::numeric_limits<double>::max());
    }
   
    VectorXd mu_vec(n_sol);
    mu_vec= (e_psinv * data->evaluations.cast<double>());
    e_mu = mu_vec.sum() / psi_sum;

    e_yon = data->evaluations.cast<double>().array() - e_mu;
    e_ssq = (e_yon.transpose() * e_psinv * e_yon) ;
    e_ssq = e_ssq / (double)n_sol;
    
  //  cout << sigmasqr << endl << endl;
    if(e_ssq < 0){
       return((std::numeric_limits<double>::max())-e_ssq);
    }
    
    double neglnlike= n_sol * log(e_ssq) + psi_det;
    return(neglnlike);
    
}


void calculateDistances(bool first, int indexAdd){
   int n_sol=solutions.rows();
   int aux;

   if(first){
       data.distance_matrix = MatrixXd::Zero(n_sol,n_sol);
       for(int i=0; i < (n_sol-1) ; i++){
           for(int j=(i+1); j<n_sol ; j++){
                aux=0;
                for(int k=0; k < (solutions.cols()-1) ; k++) {
                    if(solutions(i,k) != solutions(j,k)) aux++;
                }
                data.distance_matrix(i,j) = ((double) aux) / ((double) solutions.cols()-1.0 );
                data.distance_matrix(j,i) = data.distance_matrix(i,j);
            }
        }
   }else {
        cout << "Updating distance index: " << indexAdd << "\n";
        data.distance_matrix.conservativeResize(n_sol, Eigen::NoChange);
        data.distance_matrix.conservativeResize(Eigen::NoChange, n_sol);
        
        //Space for the new solution of is not the last one!
        if(indexAdd < (n_sol-1)){
          data.distance_matrix.block((indexAdd+1), 0, (n_sol-indexAdd-1) , n_sol) = data.distance_matrix.block(indexAdd, 0, (n_sol-indexAdd-1), n_sol).eval();
          data.distance_matrix.block(0, (indexAdd+1) , n_sol, (n_sol-indexAdd-1) ) = data.distance_matrix.block(0, indexAdd, n_sol, (n_sol-indexAdd-1)).eval();
        }
        
        for(int i=0; i<= (n_sol-1) ; i++){
           if(i!=indexAdd){
              aux=0;
              for(int k=0; k < (solutions.cols()-1) ; k++) {
                 if(solutions(i,k) != solutions(indexAdd,k)) aux++;
              }

              data.distance_matrix(i,indexAdd) = ((double) aux) / ((double) solutions.cols()-1.0);
              data.distance_matrix(indexAdd,i) = data.distance_matrix(i,indexAdd);
           }else{
              data.distance_matrix(indexAdd,indexAdd) = 0;
           }
        }
    }
    data.evaluations = evaluations;
}


bool FRLH (){
    cout << "Bulding Model...\n";
    int n_sol=solutions.rows();
    double lower_theta=log10(1e-6);
    double upper_theta=log10(1e5);
    int budget_theta=100;

    //x0 is a vector that has the initial values of the parameters
    //to estimate. (theta, p, lambda. In this case we make only theta)
    std::vector <double> x0;
    x0.push_back((double)n_sol/100.0); //value suggested by Martin, he is not sure.
    x0[0]=min(x0[0], upper_theta);
    x0[0]=max(x0[0], lower_theta);     

    bool do_continue=TRUE, model_done=FALSE;

    //Distance should be already calculated!
        
    //uses 1 because is only theta! (if not x0 should be a vector with lamba and p also)
    nlopt::opt solver(nlopt::LN_NELDERMEAD,(unsigned int)1);
    solver.set_min_objective(objective, &data);
    double f_obj;
    nlopt::result res;
    
    eval_count =0;
    eval_min= std::numeric_limits<double>::infinity();

    do{
        solver.set_lower_bounds(lower_theta);
        solver.set_upper_bounds(upper_theta);
        solver.set_ftol_rel(1e-6);
       // solver.set_ftol_abs(1e-12);
        solver.set_xtol_rel(- std::numeric_limits<double>::infinity());
        solver.set_maxeval(budget_theta);
        try{
          res = solver.optimize(x0, f_obj);
        }catch (exception& e) {
            cout << "An exception occurred. Exception  " << e.what() << '\n';
         }

        if(eval_min > f_obj ){
           eval_min=f_obj;
           eval_sol=x0[0];
           loval=lower_theta;
           upval=upper_theta;
           dmodeltheta =  pow(10, eval_sol);
           theta = x0[0];
           ssq = e_ssq;
           mu = e_mu;
           yon = e_yon;
           psi = e_psi;
           psinv = e_psinv; 
           model_done = TRUE;
        }
        //Continue looking for a better value!
        x0[0] = lower_theta + (upper_theta - lower_theta) * ran01(&seed);
        if(eval_count >= budget_theta){
           do_continue=FALSE;
        }
    }while (do_continue);

    return(model_done);

}



